import os.path as osp
from functools import partial
from typing import Any, Dict, Tuple

import mmengine
import torch
import torch.nn as nn
from bayesian_lora import calculate_kronecker_factors, cholesky_decompose_small_factors
from torch.utils.data import DataLoader
from transformers import PreTrainedTokenizer

from ..utils import Device, setup_logger


class KFACTrainer:

    def __init__(self, cfg: mmengine.Config, tokenizer: PreTrainedTokenizer, target_ids: torch.LongTensor):
        self.cfg = cfg
        self.tokenizer = tokenizer
        self.target_ids = target_ids

    def forward_call(self, model: nn.Module, batch: Any, device: Device) -> torch.Tensor:
        prompts = batch['prompts']
        inputs = self.tokenizer(prompts, **self.cfg.tokenizer_run_cfg).to(device)
        outputs = model(**inputs)
        logits = outputs.logits[:, -1, self.target_ids]
        return logits.softmax(-1)

    def compute_kfac_factors(self, model: nn.Module, data_loader: DataLoader,
                             work_dir: str) -> Dict[str, Tuple[torch.Tensor, torch.Tensor]]:
        logger = setup_logger('evidential')
        logger.info('Computing KFAC factors...')
        device = next(model.parameters()).device
        forward_call = partial(self.forward_call, device=device)

        factors = calculate_kronecker_factors(
            model,
            forward_call,
            data_loader,
            self.cfg.lalora.n_kfac,
            self.cfg.lalora.lr_threshold, ['lora'],
            use_tqdm=True)
        # Calculate Cholesky decomposition of the smaller factors
        factors = cholesky_decompose_small_factors(factors, self.cfg.lalora.lr_threshold, device, torch.float32)
        kfac_path = osp.join(work_dir, 'kronecker_factors.pth')
        torch.save({'factors': factors}, kfac_path)
        logger.info(f'Kronecker factors saved to {kfac_path}')

        return factors
