from typing import Any, Dict, Tuple, Union

import mmengine
import torch
import torch.nn.functional as F
from bayesian_lora.main import stable_cholesky
from torch import nn
from transformers import Trainer

from ..models import custom_jacobian_mean, custom_variance
from ..utils import setup_logger


class LALoraTrainer(Trainer):

    def __init__(
            self,
            cfg: mmengine.Config,
            target_ids: torch.LongTensor,
            kfac_factors: Dict[str, Tuple[torch.Tensor, torch.Tensor]],
            **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg
        self.target_ids = target_ids.to(self.args.device)
        self.kfac_factors = kfac_factors
        self.lora_params = None
        self.s2 = torch.tensor(self.cfg.lalora.prior_var, device=self.args.device)
        logger = setup_logger('evidential')
        logger.info(f'Using LA-Lora prior variance: {self.s2}')

    def _set_lalora_params(self, model: nn.Module) -> None:
        lora_params = {k: v for k, v in dict(model.named_parameters()).items() if v.requires_grad}
        # Sanity check
        for k in lora_params.keys():
            assert 'lora' in k.lower()

        self.lora_params = lora_params

    def compute_loss(self,
                     model: nn.Module,
                     inputs: Dict[str, Union[torch.Tensor, Any]],
                     return_outputs: bool = False) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        prompts = inputs.pop('prompts')
        labels = inputs.pop('labels').to(self.args.device)
        inputs = self.tokenizer(prompts, **self.cfg.tokenizer_run_cfg).to(self.args.device)

        if self.lora_params is None:
            self._set_lalora_params(model)

        jacobian, f_mu = custom_jacobian_mean(
            model,
            inputs,
            lora_params=self.lora_params,
            output_callback=lambda outputs: outputs.logits[:, -1, self.target_ids]
        )
        f_var = custom_variance(
            inputs.input_ids.shape[0],
            jacobian,
            self.kfac_factors,
            self.s2,
            self.target_ids.shape[-1],
            self.cfg.model.peft_cfg.r,
            self.cfg.lalora.n_kfac,
            self.args.device,
        )
        # Sample logits from a Gaussian parametrised by f_mu, f_var
        L = stable_cholesky(f_var)
        samples = 100_000
        f_mu = f_mu.expand(samples, *f_mu.shape)
        L = L.expand(samples, *L.shape)
        eps = torch.randn_like(f_mu).unsqueeze(-1)
        logits = f_mu[..., None] + L @ eps
        # first compute the average probabilities, and then convert them to logits
        probs = logits.squeeze(-1).softmax(-1).mean(0)
        probs = torch.clamp(probs, min=1e-10)
        ref_probs = probs[:, -1]
        logits = torch.log(probs) - torch.log(ref_probs.unsqueeze(1))

        loss = F.cross_entropy(logits, labels)

        # create a dict to mimic the output of most huggingface models, in order to be compatible with the Trainer API.
        outputs = {'loss': loss, 'logits': logits}

        return (loss, outputs) if return_outputs else loss
